import calendar
import datetime
import time
from _strptime import (_CACHE_MAX_SIZE, IGNORECASE, LocaleTime, _cache_lock,
                       _calc_julian_from_U_or_W, _getlang, _regex_cache,
                       re_compile, re_escape)
from calendar import monthrange
from collections import defaultdict
from typing import List, Optional, Set, Tuple

from kanjize import kanji2number

_ERA_DATA_COMMON, _ERA_DATA_GENERAL, _ERA_DATA_DAIKAKUJI, _ERA_DATA_JIMYOUIN = [], [], [], []

_era_kanji_dict = defaultdict(set)
_era_alphabet_dict = defaultdict(set)
_era_alphabet_vowel_shortened_dict = defaultdict(set)
_era_alphabet_head_dict = defaultdict(set)

_JAPANERA_TimeRE_cache = None


def _set_era_data(era_data_common, era_data_general, era_data_daikakuji, era_data_jimyouin):
    global _ERA_DATA_COMMON, _ERA_DATA_GENERAL, _ERA_DATA_DAIKAKUJI, _ERA_DATA_JIMYOUIN
    _ERA_DATA_COMMON = era_data_common
    _ERA_DATA_GENERAL = era_data_general
    _ERA_DATA_DAIKAKUJI = era_data_daikakuji
    _ERA_DATA_JIMYOUIN = era_data_jimyouin

    for _era in _ERA_DATA_COMMON + _ERA_DATA_GENERAL + _ERA_DATA_DAIKAKUJI + _ERA_DATA_JIMYOUIN:
        _era_kanji_dict[_era.kanji].add(_era)
        _era_alphabet_dict[_era.english].add(_era)
        _era_alphabet_vowel_shortened_dict[_era.english_vowel_shortened].add(_era)
        _era_alphabet_head_dict[_era.english_head].add(_era)

    global _JAPANERA_TimeRE_cache
    _JAPANERA_TimeRE_cache = TimeRE()


class TimeRE(dict):
    """Handle conversion from format directives to regexes."""

    def __init__(self, locale_time=None):
        """Create keys/values.

        Order of execution is important for dependency reasons.

        """
        if locale_time:
            self.locale_time = locale_time
        else:
            self.locale_time = LocaleTime()
        base = super()

        base.__init__({
            # Added for Japanera
            '-K': self.__seqToRE(_era_kanji_dict.keys(), '_K'),
            '-E': self.__seqToRE(_era_alphabet_dict.keys(), '_E'),
            '-e': self.__seqToRE(_era_alphabet_vowel_shortened_dict.keys(), '_e'),
            '-h': self.__seqToRE(_era_alphabet_head_dict.keys(), '_h'),
            '-n': r"(?P<_n>[一二三四五六七八九]?十[一二三四五六七八九]?|"
                  r"[一二三四五六七八九]|"
                  r"元)",
            '-N': r"(?P<_N>"
                  r"[一二三四五六七八九]?千([一二三四五六七八九]?百)?([一二三四五六七八九]?十)?[一二三四五六七八九]?|"
                  r"[一二三四五六七八九]?百([一二三四五六七八九]?十)?[一二三四五六七八九]?|"
                  r"[一二三四五六七八九]?十[一二三四五六七八九]?|"
                  r"[一二三四五六七八九]|"
                  r"元)",
            '-y': r"(?P<_y>\d\d|\d|"
                  r"元)",
            '-Y': r"(?P<_Y>\d{1,4}|"
                  r"元)",
            '-m': r"(?P<_m>1[0-2]|0[1-9]|[1-9]|十[一二]?|[一二三四五六七八九])",
            '-d': r"(?P<_d>3[0-1]|[1-2]\d|0[1-9]|[1-9]| [1-9]|三十一?|二?十[一二三四五六七八九]?|[一二三四五六七八九])",
            '-a': r"(?P<_a>[月火水木金土日])",

            # Default below
            # The " [1-9]" part of the regex is to make %c from ANSI C work
            'd': r"(?P<d>3[0-1]|[1-2]\d|0[1-9]|[1-9]| [1-9])",
            'f': r"(?P<f>[0-9]{1,6})",
            'H': r"(?P<H>2[0-3]|[0-1]\d|\d)",
            'I': r"(?P<I>1[0-2]|0[1-9]|[1-9])",
            'G': r"(?P<G>\d\d\d\d)",
            'j': r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|0[1-9]|[1-9])",
            'm': r"(?P<m>1[0-2]|0[1-9]|[1-9])",
            'M': r"(?P<M>[0-5]\d|\d)",
            'S': r"(?P<S>6[0-1]|[0-5]\d|\d)",
            'U': r"(?P<U>5[0-3]|[0-4]\d|\d)",
            'w': r"(?P<w>[0-6])",
            'u': r"(?P<u>[1-7])",
            'V': r"(?P<V>5[0-3]|0[1-9]|[1-4]\d|\d)",
            # W is set below by using 'U'
            'y': r"(?P<y>\d\d|\d)",
            # XXX: Does 'Y' need to worry about having less or more than
            #     4 digits?
            'Y': r"(?P<Y>\d\d\d\d)",
            'z': r"(?P<z>[+-]\d\d:?[0-5]\d(:?[0-5]\d(\.\d{1,6})?)?|Z)",
            'A': self.__seqToRE(self.locale_time.f_weekday, 'A'),
            'a': self.__seqToRE(self.locale_time.a_weekday, 'a'),
            'B': self.__seqToRE(self.locale_time.f_month[1:], 'B'),
            'b': self.__seqToRE(self.locale_time.a_month[1:], 'b'),
            'p': self.__seqToRE(self.locale_time.am_pm, 'p'),
            'Z': self.__seqToRE((tz for tz_names in self.locale_time.timezone
                                 for tz in tz_names),
                                'Z'),
            '%': '%'})
        base.__setitem__('W', base.__getitem__('U').replace('U', 'W'))
        base.__setitem__('c', self.pattern(self.locale_time.LC_date_time))
        base.__setitem__('x', self.pattern(self.locale_time.LC_date))
        base.__setitem__('X', self.pattern(self.locale_time.LC_time))

    def __seqToRE(self, to_convert, directive):
        """Convert a list to a regex string for matching a directive.

        Want possible matching values to be from longest to shortest.  This
        prevents the possibility of a match occurring for a value that also
        a substring of a larger value that should have matched (e.g., 'abc'
        matching when 'abcdef' should have been the match).

        """
        to_convert = sorted(to_convert, key=len, reverse=True)
        for value in to_convert:
            if value != '':
                break
        else:
            return ''
        regex = '|'.join(re_escape(stuff) for stuff in to_convert)
        regex = '(?P<%s>%s' % (directive, regex)
        return '%s)' % regex

    def pattern(self, format):
        """Return regex pattern for the format string.

        Need to make sure that any characters that might be interpreted as
        regex syntax are escaped.

        """
        processed_format = ''
        # The sub() call escapes all characters that might be misconstrued
        # as regex syntax.  Cannot use re.escape since we have to deal with
        # format directives (%m, etc.).
        regex_chars = re_compile(r"([\\.^$*+?\(\){}\[\]|])")
        format = regex_chars.sub(r"\\\1", format)
        whitespace_replacement = re_compile(r'\s+')
        format = whitespace_replacement.sub(r'\\s+', format)
        while '%' in format:
            directive_index = format.index('%') + 1
            if format[directive_index] == '-':
                directive = format[directive_index:directive_index + 2]
            else:
                directive = format[directive_index]
            processed_format = "%s%s%s" % (processed_format,
                                           format[:directive_index - 1],
                                           self[directive])
            format = format[directive_index + len(directive):]
        return "%s%s" % (processed_format, format)

    def compile(self, format):
        """Return a compiled re object for the format string."""
        return re_compile(self.pattern(format), IGNORECASE)


def _calc_julian_from_V(iso_year, iso_week, iso_weekday):
    """Calculate the Julian day based on the ISO 8601 year, week, and weekday.
    ISO weeks start on Mondays, with week 01 being the week containing 4 Jan.
    ISO week days range from 1 (Monday) to 7 (Sunday).
    """
    correction = datetime_date(iso_year, 1, 4).isoweekday() + 3
    ordinal = (iso_week * 7) + iso_weekday - correction
    # ordinal may be negative or 0 now, which means the date is in the previous
    # calendar year
    if ordinal < 1:
        ordinal += datetime_date(iso_year, 1, 1).toordinal()
        iso_year -= 1
        ordinal -= datetime_date(iso_year, 1, 1).toordinal()
    return iso_year, ordinal


def _strptime(data_string, format="%a %b %d %H:%M:%S %Y"):
    """Return a 2-tuple consisting of a time struct and an int containing
    the number of microseconds based on the input string and the
    format string."""

    for index, arg in enumerate([data_string, format]):
        if not isinstance(arg, str):
            msg = "strptime() argument {} must be str, not {}"
            raise TypeError(msg.format(index, type(arg)))

    with _cache_lock:
        global _JAPANERA_TimeRE_cache
        locale_time = _JAPANERA_TimeRE_cache.locale_time
        if (_getlang() != locale_time.lang or
                time.tzname != locale_time.tzname or
                time.daylight != locale_time.daylight):
            _JAPANERA_TimeRE_cache = TimeRE()
            _regex_cache.clear()
            locale_time = _JAPANERA_TimeRE_cache.locale_time
        if len(_regex_cache) > _CACHE_MAX_SIZE:
            _regex_cache.clear()
        format_regex = _regex_cache.get(format)
        if not format_regex:
            try:
                format_regex = _JAPANERA_TimeRE_cache.compile(format)
            # KeyError raised when a bad format is found; can be specified as
            # \\, in which case it was a stray % but with a space after it
            except KeyError as err:
                bad_directive = err.args[0]
                if bad_directive == "\\":
                    bad_directive = "%"
                del err
                raise ValueError("'%s' is a bad directive in format '%s'" %
                                 (bad_directive, format)) from None
            # IndexError only occurs when the format string is "%"
            except IndexError:
                raise ValueError("stray %% in format '%s'" % format) from None
            _regex_cache[format] = format_regex
    found = format_regex.match(data_string)
    if not found:
        raise ValueError("time data %r does not match format %r" %
                         (data_string, format))
    if len(data_string) != found.end():
        raise ValueError("unconverted data remains: %s" %
                         data_string[found.end():])

    iso_year = year = None
    month = day = None
    hour = minute = second = fraction = 0
    tz = -1
    gmtoff = None
    gmtoff_fraction = 0
    # Default to -1 to signify that values not known; not critical to have,
    # though
    iso_week = week_of_year = None
    week_of_year_start = None
    # weekday and julian defaulted to None so as to signal need to calculate
    # values
    weekday = julian = None
    found_dict = found.groupdict()

    era_kanji = era_english = era_english_vowel_shortened = era_head = None
    relative_year = None

    for group_key in found_dict.keys():
        # Directives not explicitly handled below:
        #   c, x, X
        #      handled by making out of other directives
        #   U, W
        #      worthless without day of the week
        #  ===== ADDED BY JAPANERA BELOW =====
        if group_key == '_K':
            era_kanji = found_dict['_K']
        elif group_key == '_E':
            era_english = found_dict['_E']
        elif group_key == '_e':
            era_english_vowel_shortened = found_dict['_e']
        elif group_key == '_h':
            era_head = found_dict['_h']
        elif group_key == '_n':
            found_relative_year = found_dict['_n']

            if found_relative_year == '元':
                relative_year = 1
            else:
                relative_year = kanji2number(found_relative_year)
        elif group_key == '_N':
            found_absolute_year = found_dict['_N']

            if found_absolute_year == '元':
                year = 1
            else:
                year = kanji2number(found_absolute_year)
        elif group_key == '_y':
            found_relative_year = found_dict['_y']

            if found_relative_year == '元':
                relative_year = 1
            else:
                relative_year = int(found_relative_year)
        elif group_key == '_Y':
            found_absolute_year = found_dict['_Y']

            if found_absolute_year == '元':
                year = 1
            else:
                year = kanji2number(found_absolute_year)
        elif group_key == '_m':
            month = kanji2number(found_dict['_m'])
        elif group_key == '_d':
            day = kanji2number(found_dict['_d'])
        elif group_key == '_a':
            weekday = '月火水木金土日'.index(found_dict['_a'])
        #  ===== ADDED BY JAPANERA ABOVE =====

        elif group_key == 'y':
            year = int(found_dict['y'])
            # Open Group specification for strptime() states that a %y
            # value in the range of [00, 68] is in the century 2000, while
            # [69,99] is in the century 1900
            if year <= 68:
                year += 2000
            else:
                year += 1900
        elif group_key == 'Y':
            year = int(found_dict['Y'])
        elif group_key == 'G':
            iso_year = int(found_dict['G'])
        elif group_key == 'm':
            month = int(found_dict['m'])
        elif group_key == 'B':
            month = locale_time.f_month.index(found_dict['B'].lower())
        elif group_key == 'b':
            month = locale_time.a_month.index(found_dict['b'].lower())
        elif group_key == 'd':
            day = int(found_dict['d'])
        elif group_key == 'H':
            hour = int(found_dict['H'])
        elif group_key == 'I':
            hour = int(found_dict['I'])
            ampm = found_dict.get('p', '').lower()
            # If there was no AM/PM indicator, we'll treat this like AM
            if ampm in ('', locale_time.am_pm[0]):
                # We're in AM so the hour is correct unless we're
                # looking at 12 midnight.
                # 12 midnight == 12 AM == hour 0
                if hour == 12:
                    hour = 0
            elif ampm == locale_time.am_pm[1]:
                # We're in PM so we need to add 12 to the hour unless
                # we're looking at 12 noon.
                # 12 noon == 12 PM == hour 12
                if hour != 12:
                    hour += 12
        elif group_key == 'M':
            minute = int(found_dict['M'])
        elif group_key == 'S':
            second = int(found_dict['S'])
        elif group_key == 'f':
            s = found_dict['f']
            # Pad to always return microseconds.
            s += "0" * (6 - len(s))
            fraction = int(s)
        elif group_key == 'A':
            weekday = locale_time.f_weekday.index(found_dict['A'].lower())
        elif group_key == 'a':
            weekday = locale_time.a_weekday.index(found_dict['a'].lower())
        elif group_key == 'w':
            weekday = int(found_dict['w'])
            if weekday == 0:
                weekday = 6
            else:
                weekday -= 1
        elif group_key == 'u':
            weekday = int(found_dict['u'])
            weekday -= 1
        elif group_key == 'j':
            julian = int(found_dict['j'])
        elif group_key in ('U', 'W'):
            week_of_year = int(found_dict[group_key])
            if group_key == 'U':
                # U starts week on Sunday.
                week_of_year_start = 6
            else:
                # W starts week on Monday.
                week_of_year_start = 0
        elif group_key == 'V':
            iso_week = int(found_dict['V'])
        elif group_key == 'z':
            z = found_dict['z']
            if z == 'Z':
                gmtoff = 0
            else:
                if z[3] == ':':
                    z = z[:3] + z[4:]
                    if len(z) > 5:
                        if z[5] != ':':
                            msg = f"Inconsistent use of : in {found_dict['z']}"
                            raise ValueError(msg)
                        z = z[:5] + z[6:]
                hours = int(z[1:3])
                minutes = int(z[3:5])
                seconds = int(z[5:7] or 0)
                gmtoff = (hours * 60 * 60) + (minutes * 60) + seconds
                gmtoff_remainder = z[8:]
                # Pad to always return microseconds.
                gmtoff_remainder_padding = "0" * (6 - len(gmtoff_remainder))
                gmtoff_fraction = int(gmtoff_remainder + gmtoff_remainder_padding)
                if z.startswith("-"):
                    gmtoff = -gmtoff
                    gmtoff_fraction = -gmtoff_fraction
        elif group_key == 'Z':
            # Since -1 is default value only need to worry about setting tz if
            # it can be something other than -1.
            found_zone = found_dict['Z'].lower()
            for value, tz_values in enumerate(locale_time.timezone):
                if found_zone in tz_values:
                    # Deal with bad locale setup where timezone names are the
                    # same and yet time.daylight is true; too ambiguous to
                    # be able to tell what timezone has daylight savings
                    if (time.tzname[0] == time.tzname[1] and
                            time.daylight and found_zone not in ("utc", "gmt")):
                        break
                    else:
                        tz = value
                        break
    # Deal with the cases where ambiguities arize
    # don't assume default values for ISO week/year
    if year is None and iso_year is not None:
        if iso_week is None or weekday is None:
            raise ValueError("ISO year directive '%G' must be used with "
                             "the ISO week directive '%V' and a weekday "
                             "directive ('%A', '%a', '%w', or '%u').")
        if julian is not None:
            raise ValueError("Day of the year directive '%j' is not "
                             "compatible with ISO year directive '%G'. "
                             "Use '%Y' instead.")
    elif week_of_year is None and iso_week is not None:
        if weekday is None:
            raise ValueError("ISO week directive '%V' must be used with "
                             "the ISO year directive '%G' and a weekday "
                             "directive ('%A', '%a', '%w', or '%u').")
        else:
            raise ValueError("ISO week directive '%V' is incompatible with "
                             "the year directive '%Y'. Use the ISO year '%G' "
                             "instead.")
    tzname = found_dict.get("Z")

    # If we know the week of the year and what day of that week, we can figure
    # out the Julian day of the year.
    if julian is None and weekday is not None:
        if week_of_year is not None:
            week_starts_Mon = True if week_of_year_start == 0 else False
            julian = _calc_julian_from_U_or_W(year, week_of_year, weekday,
                                              week_starts_Mon)
        elif iso_year is not None and iso_week is not None:
            year, julian = _calc_julian_from_V(iso_year, iso_week, weekday + 1)
        if julian is not None and julian <= 0:
            year -= 1
            yday = 366 if calendar.isleap(year) else 365
            julian += yday
    return (era_kanji, era_english, era_english_vowel_shortened,
            era_head, relative_year), \
        (year, month, day,
         hour, minute, second,
         weekday, julian, tz, tzname,
         gmtoff), fraction, gmtoff_fraction


def find_era_and_date(era_kanji: Optional[str] = None,
                      era_english: Optional[str] = None,
                      era_english_vowel_shortened: Optional[str] = None,
                      era_head_english: Optional[str] = None,
                      absolute_year: Optional[int] = None,
                      relative_year: Optional[int] = None,
                      month: Optional[int] = None,
                      day: Optional[int] = None,
                      allow_date_after_end_of_era: bool = False,
                      ) -> List[Tuple["Era", datetime.date]]:
    """
    Find era and date from given information.
    Args:
        era_kanji: Kanji of era name
        era_english: English of era name
        era_english_vowel_shortened: English of era name with vowel shortened
        era_head_english: English of era name with only first letter
        absolute_year: Absolute year
        relative_year: Relative year from era
        month: Month
        day: Day
        allow_date_after_end_of_era: If True, allow date after end of era

    Returns: List of era and date

    """
    era_set = None

    for text, _dict in zip((era_kanji, era_english, era_english_vowel_shortened, era_head_english),
                           (_era_kanji_dict, _era_alphabet_dict, _era_alphabet_vowel_shortened_dict,
                            _era_alphabet_head_dict)):
        if text:
            _found = _dict[text]
            era_set = era_set & _found if era_set else _found
    if absolute_year is not None:
        _found = find_eras_with_year(absolute_year)
        era_set = era_set & _found if era_set else _found
    if (era_kanji or era_english or era_english_vowel_shortened or
        era_head_english or absolute_year is not None) and not era_set:
        raise ValueError("Era_ information given but no match era found.")

    era_set = era_set or _ERA_DATA_GENERAL + _ERA_DATA_COMMON + _ERA_DATA_DAIKAKUJI + _ERA_DATA_JIMYOUIN
    era_list = sorted(era_set, key=lambda x: (x.since, x.era_type.value))

    result = []
    for era in era_list:
        dt = era.since
        if absolute_year is not None and absolute_year != dt.year:
            # if absolute_year is different from era's since year, we can make result new years day
            dt = datetime.date(absolute_year, 1, 1)
        elif relative_year is not None and relative_year > 1:
            # relative_year starts from 1, so we need to add 1 to relative_year
            dt = datetime.date(dt.year + relative_year - 1, 1, 1)
        if month is not None:
            if dt.year != era.since.year or month != dt.month:
                # if month is specified to be different from `since` month, 1st of the month is most early date
                dt = dt.replace(month=month, day=1)
                if absolute_year is None and relative_year is None and dt < era.since:
                    dt = dt.replace(year=dt.year + 1)
            if day is not None:  # if month is specified, we can't use loop below to find the month
                if month == 2 and day == 29:
                    if absolute_year is None and relative_year is None:
                        # if month is February and day is 29 and no year is specified, we need to find next leap year
                        dt = dt.replace(year=find_closest_leap_year(dt.year))
                try:
                    dt = dt.replace(day=day)
                except ValueError:  # out of range
                    continue
        elif day is not None:
            while True:  # we know `day` is not more than 31 because of legit regex
                try:
                    dt = dt.replace(day=day)
                    break
                except ValueError:  # out of range
                    days_in_month = monthrange(dt.year, dt.month)[1]
                    dt += datetime.timedelta(days=days_in_month)

        if allow_date_after_end_of_era:
            if dt < era.since:
                continue
        elif dt not in era:
            continue
        result.append((era, dt))
    return result


def find_closest_leap_year(year: int) -> int:
    """
    Find the closest leap year from `year`. If `year` is leap year, return `year`.
    Args:
        year: year to find

    Returns: closest leap year from `year`

    """
    if not year % 100 and year % 400:
        return year + 4
    if not year % 4:
        return year
    if (year % 100 > 96) and (year % 400 <= 396):
        return year + (8 - year % 4)
    return year + (4 - year % 4)


def find_eras_with_year(year: int) -> Set["Era"]:
    """
    Find all eras that contains `year`.
    Args:
        year: year to find

    Returns: set of Era that contains `year`
    """

    def _find_first_era_after_year_index(era_list: List["Era"]) -> int:
        # return the index of first era that starts after `year`
        ok = len(era_list)
        ng = -1
        while abs(ok - ng) > 1:
            mid = (ok + ng) // 2
            if era_list[mid].since.year <= year:
                ng = mid
            else:
                ok = mid
        return ok

    result = set(_ERA_DATA_COMMON)
    for era_list in (_ERA_DATA_GENERAL, _ERA_DATA_DAIKAKUJI, _ERA_DATA_JIMYOUIN):
        for i in range(_find_first_era_after_year_index(era_list) - 1, -1, -1):
            era = era_list[i]
            # we know that this era must be started before or exact `year`
            # so we can check the `era.until.year` to see if it contains
            # `year`. and notice that current `era`'s until can be None
            if era.until and year > era.until.year:
                break
            result.add(era)

    return result



# -*- coding: utf-8 -*-
import datetime
import re
from typing import Optional, List, Union
from warnings import warn

from kanjize import number2kanji

from .era_data import EraType, _ERA_DATA_COMMON, _ERA_DATA_GENERAL, _ERA_DATA_DAIKAKUJI, _ERA_DATA_JIMYOUIN
from .parser import _strptime, find_era_and_date, _set_era_data


class Era:
    def __init__(self, kanji: Optional[str], english: Optional[str], since: datetime.date,
                 until: Optional[datetime.date], era_type: EraType):
        self._kanji = kanji
        self._english = english
        self.since = since
        self.until = until
        self.era_type = era_type

    @property
    def kanji(self) -> str:
        return self._kanji or "不明"

    @property
    def english(self) -> str:
        return self._english or "Unknown"

    @property
    def english_vowel_shortened(self) -> str:
        return self.english.lower().replace("ou", "o").replace("uu", "u").title()  # much faster than regex

    @property
    def english_head(self) -> str:
        return self.english[0]

    def relative_year_to_absolute_year(self, relative_year: int) -> int:
        return self.since.year + relative_year - 1

    def absolute_year_to_relative_year(self, absolute_year: int) -> int:
        return absolute_year - self.since.year + 1

    def calc_relative_year(self, date: datetime.date) -> int:
        return self.absolute_year_to_relative_year(date.year)

    def strftime(self, dtt: Union[datetime.date, datetime.datetime], format: str) -> str:
        relative_year = dtt.year - self.since.year + 1
        rep = {"%-K": self.kanji, "%-E": self.english, "%-e": self.english_vowel_shortened,
               "%-h": self.english_head,
               "%-n": "元" if relative_year == 1 else number2kanji(relative_year % 100),
               "%-N": "元" if relative_year == 1 else number2kanji(relative_year),
               "%-y": "{:02}".format(relative_year % 100),
               "%-Y": str(relative_year),
               "%-m": number2kanji(dtt.month),
               "%-d": number2kanji(dtt.day),
               "%-a": '月火水木金土日'[dtt.weekday()]}

        rep = dict((re.escape(k), str(v)) for k, v in rep.items())
        pattern = re.compile("|".join(rep.keys()))
        return datetime.datetime.strftime(dtt, pattern.sub(lambda m: rep[re.escape(m.group(0))], format))

    def strptime(self, date_string: str, format: str) -> "EraDateTime":
        (era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year), \
        (year, month, day, hour, minute, second, weekday, julian, tz, tzname, gmtoff), \
        fraction, gmtoff_fraction = _strptime(date_string, format)
        era_and_dates = find_era_and_date(self.kanji, self.english, None, None, year,
                                          relative_year, month, day, True)
        if not era_and_dates:
            raise ValueError("EraDate not found")

        tz = None
        if gmtoff is not None:
            tzdelta = datetime.timedelta(seconds=gmtoff, microseconds=gmtoff_fraction)
            if tzname:
                tz = datetime.timezone(tzdelta, tzname)
            else:
                tz = datetime.timezone(tzdelta)
        for era, date in era_and_dates:
            if era == self:
                return EraDateTime(date.year, date.month, date.day, hour, minute, second, fraction, tzinfo=tz, era=self)

    def __gt__(self, other):
        since = self.since or datetime.date.min
        if isinstance(other, datetime.datetime):
            return since > other.date()
        elif isinstance(other, datetime.date):
            return since > other
        elif isinstance(other, Era):
            return since > other.since
        raise TypeError(f"unsupported operand type(s) for >: '{type(self).__name__}' and '{type(other).__name__}'")

    def __lt__(self, other):
        until = self.until or datetime.date.max
        if isinstance(other, datetime.datetime):
            return until <= other.date()
        elif isinstance(other, datetime.date):
            return until <= other
        elif isinstance(other, Era):
            return until <= other.since
        raise TypeError(f"unsupported operand type(s) for <: '{type(self).__name__}' and '{type(other).__name__}'")

    def __ge__(self, other):
        since = self.since or datetime.date.min
        if isinstance(other, datetime.datetime):
            return since >= other.date()
        elif isinstance(other, datetime.date):
            return since >= other
        elif isinstance(other, Era):
            return since >= other.since
        raise TypeError(f"unsupported operand type(s) for >=: '{type(self).__name__}' and '{type(other).__name__}'")

    def __le__(self, other):
        return self < other

    def __contains__(self, item: Union[datetime.date, datetime.datetime]):
        since = self.since or datetime.date.min
        until = self.until or datetime.date.max
        if isinstance(item, datetime.datetime):
            return since <= item.date() < until
        elif isinstance(item, datetime.date):
            return since <= item < until
        else:
            raise TypeError(f"unsupported operand type(s) for in: '{type(self).__name__}' and '{type(item).__name__}'")

    def __eq__(self, other):
        if not isinstance(other, Era):
            return False
        return self.kanji == other.kanji and self.english == other.english and self.since == other.since and self.until == other.until and self.era_type == other.era_type

    def __hash__(self):
        return hash((self.kanji, self.english, self.since, self.until, self.era_type))

    def __repr__(self):
        return "Era({!r}, {!r}, {!r}, {!r}, {!r})".format(self.kanji, self.english, self.since, self.until,
                                                          self.era_type)


ERA_DATA_COMMON = [Era(kanji, english, since, until, _type) for kanji, english, since, until, _type in
                   _ERA_DATA_COMMON]
ERA_DATA_GENERAL = [Era(kanji, english, since, until, _type) for kanji, english, since, until, _type in
                    _ERA_DATA_GENERAL]
ERA_DATA_DAIKAKUJI = [Era(kanji, english, since, until, _type) for kanji, english, since, until, _type in
                      _ERA_DATA_DAIKAKUJI]
ERA_DATA_JIMYOUIN = [Era(kanji, english, since, until, _type) for kanji, english, since, until, _type in
                     _ERA_DATA_JIMYOUIN]

_set_era_data(ERA_DATA_COMMON, ERA_DATA_GENERAL, ERA_DATA_DAIKAKUJI, ERA_DATA_JIMYOUIN)


class EraDate(datetime.date):
    def __new__(cls, year: int, month: Optional[int] = None, day: Optional[int] = None, era: Optional[Era] = None):
        self = super().__new__(cls, year, month, day)
        if era:
            self.era = era
        else:
            result = find_era_and_date(absolute_year=year, month=month, day=day)
            if not result:
                raise ValueError("Era not found")  # Maybe this can't be happened because of Common Era
            self.era = result[-1][0]
        if self not in self.era:
            warn("Date is not in era", RuntimeWarning)
        return self

    @classmethod
    def strptime(cls, date_string: str, format: str, allow_date_after_end_of_era=False) -> List["EraDate"]:
        (era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year), \
        (year, month, day, hour, minute, second, weekday, julian, tz, tzname, gmtoff), \
        fraction, gmtoff_fraction = _strptime(date_string, format)
        era_and_dates = find_era_and_date(era_kanji, era_english, era_english_vowel_shortened, era_head, year,
                                          relative_year, month, day, allow_date_after_end_of_era)
        if not era_and_dates:
            raise ValueError("EraDate not found")
        return [cls.from_date(date, era=era) for era, date in era_and_dates]

    def strftime(self, format: str) -> str:
        """
        %-K: Kanji era name
        %-E: English era name
        %-e: English era name vowel shortened
        %-h: Head of english era name
        %-n: Up to two digit relative year of corresponding era in Kanji. But return "元" for the first year
        %-N: Up to four digit relative year of corresponding era in Kanji. But return "元" for the first year
        %-y: Two digit relative year of corresponding era.
        %-Y: Up to four digit relative year of corresponding era.
        %-m: Month of date in Kanji
        %-d: Day of date in Kanji
        %-a: Weekday of date in Kanji
        + datetime.strftime's format
        """
        relative_year = self.year - self.era.since.year + 1
        rep = {"%-K": self.era.kanji, "%-E": self.era.english, "%-e": self.era.english_vowel_shortened,
               "%-h": self.era.english_head,
               "%-n": "元" if relative_year == 1 else number2kanji(relative_year % 100),
               "%-N": "元" if relative_year == 1 else number2kanji(relative_year),
               "%-y": "{:02}".format(relative_year % 100),
               "%-Y": str(relative_year),
               "%-m": number2kanji(self.month),
               "%-d": number2kanji(self.day),
               "%-a": '月火水木金土日'[self.weekday()]}

        rep = dict((re.escape(k), str(v)) for k, v in rep.items())
        pattern = re.compile("|".join(rep.keys()))
        return datetime.datetime.strftime(self, pattern.sub(lambda m: rep[re.escape(m.group(0))], format))

    @classmethod
    def from_date(cls, dt: datetime.date, era: Optional[Era] = None) -> "EraDate":
        if not era:
            result = find_era_and_date(absolute_year=dt.year, month=dt.month, day=dt.day)
            if not result:
                raise ValueError("Era not found")  # Maybe this can't be happened because of Common Era
            era = result[-1][0]
        return cls(year=dt.year, month=dt.month, day=dt.day, era=era)

    @classmethod
    def list_from_date(cls, dt: datetime.date, eras: Optional[List[Era]] = None) -> List["EraDate"]:
        if not eras:
            result = find_era_and_date(absolute_year=dt.year, month=dt.month, day=dt.day)
            if not result:
                raise ValueError("Era not found")  # Maybe this can't be happened because of Common Era
            eras = [era for era, date in result]
        return [cls(year=dt.year, month=dt.month, day=dt.day, era=era) for era in eras]

    def to_date(self) -> datetime.date:
        return datetime.date(year=self.year, month=self.month, day=self.day)

    def __hash__(self):
        return hash((self.year, self.month, self.day, self.era))

    def __eq__(self, other):
        if isinstance(other, EraDate):
            return self.year == other.year and self.month == other.month and self.day == other.day and self.era == other.era
        else:
            return super().__eq__(other)

    def __repr__(self):
        return "EraDate({!r}, {!r}, {!r}, {!r})".format(self.year, self.month, self.day, self.era)

    def __str__(self):
        return self.strftime("%-K%-y年 %m月%d日")


class EraDateTime(EraDate, datetime.datetime):
    def __new__(cls, year: int, month: Optional[int] = None, day: Optional[int] = None, hour: Optional[int] = 0,
                minute: int = 0, second: int = 0, microsecond: int = 0, tzinfo: Optional[datetime.tzinfo] = None, *,
                fold: int = 0, era: Optional[Era] = None):
        self = datetime.datetime.__new__(cls, year=year, month=month, day=day, hour=hour, minute=minute, second=second,
                                         microsecond=microsecond, tzinfo=tzinfo, fold=fold)

        if era:
            self.era = era
        else:
            result = find_era_and_date(absolute_year=year, month=month, day=day)
            if not result:
                raise ValueError("Era not found")  # Maybe this can't be happened because of Common Era
            self.era = result[-1][0]
        if self not in self.era:
            warn("Date is not in era", RuntimeWarning)
        return self

    @classmethod
    def strptime(cls, date_string: str, format: str, allow_date_after_end_of_era=False) -> List["EraDateTime"]:
        (era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year), \
        (year, month, day, hour, minute, second, weekday, julian, tz, tzname, gmtoff), \
        fraction, gmtoff_fraction = _strptime(date_string, format)
        era_and_dates = find_era_and_date(era_kanji, era_english, era_english_vowel_shortened, era_head, year,
                                          relative_year, month, day, allow_date_after_end_of_era)
        if not era_and_dates:
            raise ValueError("EraDate not found")

        tz = None
        if gmtoff is not None:
            tzdelta = datetime.timedelta(seconds=gmtoff, microseconds=gmtoff_fraction)
            if tzname:
                tz = datetime.timezone(tzdelta, tzname)
            else:
                tz = datetime.timezone(tzdelta)
        return [cls(date.year, date.month, date.day, hour, minute, second, fraction, tzinfo=tz, era=era) for era, date
                in era_and_dates]

    @classmethod
    def list_from_datetime(cls, dtt: datetime.datetime, eras: Optional[List[Era]] = None) -> List["EraDateTime"]:
        if not eras:
            result = find_era_and_date(absolute_year=dtt.year, month=dtt.month, day=dtt.day)
            if not result:
                raise ValueError("Era not found")  # Maybe this can't be happened because of Common Era
            eras = [era for era, date in result]
        return [cls(year=dtt.year, month=dtt.month, day=dtt.day, hour=dtt.hour, minute=dtt.minute, second=dtt.second,
                    microsecond=dtt.microsecond, tzinfo=dtt.tzinfo, fold=dtt.fold, era=era) for era in eras]

    @classmethod
    def from_datetime(cls, dtt: datetime.datetime, era: Optional[Era] = None) -> "EraDateTime":
        if not era:
            result = find_era_and_date(absolute_year=dtt.year, month=dtt.month, day=dtt.day)
            if not result:
                raise ValueError("Era not found")  # Maybe this can't be happened because of Common Era
            era = result[-1][0]
        return cls(year=dtt.year, month=dtt.month, day=dtt.day, hour=dtt.hour, minute=dtt.minute, second=dtt.second,
                   microsecond=dtt.microsecond, tzinfo=dtt.tzinfo, fold=dtt.fold, era=era)

    def to_datetime(self) -> datetime.datetime:
        return datetime.datetime(year=self.year, month=self.month, day=self.day, hour=self.hour, minute=self.minute,
                                 second=self.second, microsecond=self.microsecond, tzinfo=self.tzinfo, fold=self.fold)

    def __eq__(self, other):
        if isinstance(other, EraDateTime):
            return self.year == other.year and self.month == other.month and self.day == other.day and self.hour == \
                   other.hour and self.minute == other.minute and self.second == other.second and self.microsecond == \
                   other.microsecond and self.tzinfo == other.tzinfo and self.era == other.era
        else:
            return super().__eq__(other)

    def __hash__(self):
        return hash((
            self.year, self.month, self.day, self.hour, self.minute, self.second, self.microsecond, self.tzinfo,
            self.era))

    def __repr__(self):
        return "EraDatetime({!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r})".format(self.year, self.month,
                                                                                          self.day, self.hour,
                                                                                          self.minute, self.second,
                                                                                          self.microsecond, self.tzinfo,
                                                                                          self.era)

    def __str__(self):
        return self.strftime("%-K%-y年 %m月%d日 %H時%M分%S秒")



import datetime
import enum


class EraType(enum.Enum):
    GENERAL = 'general'
    DAIKAKUJI = 'daikakuji'
    JIMYOUIN = 'jimyouin'
    COMMON = 'common'


_ERA_DATA_COMMON = [
    ("西暦", "Seireki", datetime.date(1, 1, 1), None, EraType.COMMON),
]

_ERA_DATA_GENERAL = [
    ("大化", "Taika", datetime.date(645, 7, 20), datetime.date(650, 3, 25), EraType.GENERAL),
    ("白雉", "Hakuchi", datetime.date(650, 3, 25), datetime.date(654, 11, 27), EraType.GENERAL),
    (None, None, datetime.date(654, 11, 27), datetime.date(686, 8, 17), EraType.GENERAL),
    ("朱鳥", "Shuchou", datetime.date(686, 8, 17), datetime.date(686, 10, 4), EraType.GENERAL),
    (None, None, datetime.date(686, 10, 4), datetime.date(701, 5, 7), EraType.GENERAL),
    ("大宝", "Taihou", datetime.date(701, 5, 7), datetime.date(704, 6, 20), EraType.GENERAL),
    ("慶雲", "Keiun", datetime.date(704, 6, 20), datetime.date(708, 2, 11), EraType.GENERAL),
    ("和銅", "Wadou", datetime.date(708, 2, 11), datetime.date(715, 10, 7), EraType.GENERAL),
    ("霊亀", "Reiki", datetime.date(715, 10, 7), datetime.date(717, 12, 28), EraType.GENERAL),
    ("養老", "Yourou", datetime.date(717, 12, 28), datetime.date(724, 3, 7), EraType.GENERAL),
    ("神亀", "Jinki", datetime.date(724, 3, 7), datetime.date(729, 9, 6), EraType.GENERAL),
    ("天平", "Tempyou", datetime.date(729, 9, 6), datetime.date(749, 5, 8), EraType.GENERAL),
    ("天平感宝", "TempyouKampou", datetime.date(749, 5, 8), datetime.date(749, 8, 23), EraType.GENERAL),
    ("天平勝宝", "TempyouSyouhou", datetime.date(749, 8, 23), datetime.date(757, 9, 10), EraType.GENERAL),
    ("天平宝字", "TempyouHouji", datetime.date(757, 9, 10), datetime.date(765, 2, 5), EraType.GENERAL),
    ("天平神護", "TempyouJingo", datetime.date(765, 2, 5), datetime.date(767, 9, 17), EraType.GENERAL),
    ("神護景雲", "JingoKeiun", datetime.date(767, 9, 17), datetime.date(770, 10, 27), EraType.GENERAL),
    ("宝亀", "Houki", datetime.date(770, 10, 27), datetime.date(781, 2, 3), EraType.GENERAL),
    ("天応", "Tennou", datetime.date(781, 2, 3), datetime.date(782, 10, 4), EraType.GENERAL),
    ("延暦", "Enryaku", datetime.date(782, 10, 4), datetime.date(806, 6, 12), EraType.GENERAL),
    ("大同", "Daidou", datetime.date(806, 6, 12), datetime.date(810, 10, 24), EraType.GENERAL),
    ("弘仁", "Kounin", datetime.date(810, 10, 24), datetime.date(824, 2, 12), EraType.GENERAL),
    ("天長", "Tenchou", datetime.date(824, 2, 12), datetime.date(834, 2, 18), EraType.GENERAL),
    ("承和", "Jouwa", datetime.date(834, 2, 18), datetime.date(848, 7, 20), EraType.GENERAL),
    ("嘉祥", "Kashou", datetime.date(848, 7, 20), datetime.date(851, 6, 5), EraType.GENERAL),
    ("仁寿", "Ninju", datetime.date(851, 6, 5), datetime.date(854, 12, 27), EraType.GENERAL),
    ("斉衡", "Saikou", datetime.date(854, 12, 27), datetime.date(857, 3, 24), EraType.GENERAL),
    ("天安", "Tennan", datetime.date(857, 3, 24), datetime.date(859, 5, 24), EraType.GENERAL),
    ("貞観", "Jougan", datetime.date(859, 5, 24), datetime.date(877, 6, 5), EraType.GENERAL),
    ("元慶", "Gangyou", datetime.date(877, 6, 5), datetime.date(885, 3, 15), EraType.GENERAL),
    ("仁和", "Ninna", datetime.date(885, 3, 15), datetime.date(889, 6, 3), EraType.GENERAL),
    ("寛平", "Kampyou", datetime.date(889, 6, 3), datetime.date(898, 5, 24), EraType.GENERAL),
    ("昌泰", "Syoutai", datetime.date(898, 5, 24), datetime.date(901, 9, 5), EraType.GENERAL),
    ("延喜", "Engi", datetime.date(901, 9, 5), datetime.date(923, 6, 3), EraType.GENERAL),
    ("延長", "Enchou", datetime.date(923, 6, 3), datetime.date(931, 5, 21), EraType.GENERAL),
    ("承平", "Jouhei", datetime.date(931, 5, 21), datetime.date(938, 6, 27), EraType.GENERAL),
    ("天慶", "Tengyou", datetime.date(938, 6, 27), datetime.date(947, 5, 20), EraType.GENERAL),
    ("天暦", "Tenryaku", datetime.date(947, 5, 20), datetime.date(957, 11, 26), EraType.GENERAL),
    ("天徳", "Tentoku", datetime.date(957, 11, 26), datetime.date(961, 3, 10), EraType.GENERAL),
    ("応和", "Ouwa", datetime.date(961, 3, 10), datetime.date(964, 8, 24), EraType.GENERAL),
    ("康保", "Kouhou", datetime.date(964, 8, 24), datetime.date(968, 9, 13), EraType.GENERAL),
    ("安和", "Anna", datetime.date(968, 9, 13), datetime.date(970, 5, 8), EraType.GENERAL),
    ("天禄", "Tenroku", datetime.date(970, 5, 8), datetime.date(974, 1, 21), EraType.GENERAL),
    ("天延", "Tenen", datetime.date(974, 1, 21), datetime.date(976, 8, 16), EraType.GENERAL),
    ("貞元", "Jougen", datetime.date(976, 8, 16), datetime.date(979, 1, 5), EraType.GENERAL),
    ("天元", "Tengen", datetime.date(979, 1, 5), datetime.date(983, 6, 3), EraType.GENERAL),
    ("永観", "Eikan", datetime.date(983, 6, 3), datetime.date(985, 5, 24), EraType.GENERAL),
    ("寛和", "Kanna", datetime.date(985, 5, 24), datetime.date(987, 5, 10), EraType.GENERAL),
    ("永延", "Eien", datetime.date(987, 5, 10), datetime.date(989, 9, 15), EraType.GENERAL),
    ("永祚", "Eiso", datetime.date(989, 9, 15), datetime.date(990, 12, 1), EraType.GENERAL),
    ("正暦", "Syouryaku", datetime.date(990, 12, 1), datetime.date(995, 3, 30), EraType.GENERAL),
    ("長徳", "Choutoku", datetime.date(995, 3, 30), datetime.date(999, 2, 6), EraType.GENERAL),
    ("長保", "Chouhou", datetime.date(999, 2, 6), datetime.date(1004, 8, 14), EraType.GENERAL),
    ("寛弘", "Kankou", datetime.date(1004, 8, 14), datetime.date(1013, 2, 14), EraType.GENERAL),
    ("長和", "Chouwa", datetime.date(1013, 2, 14), datetime.date(1017, 5, 27), EraType.GENERAL),
    ("寛仁", "Kannin", datetime.date(1017, 5, 27), datetime.date(1021, 3, 23), EraType.GENERAL),
    ("治安", "Jian", datetime.date(1021, 3, 23), datetime.date(1024, 8, 25), EraType.GENERAL),
    ("万寿", "Manju", datetime.date(1024, 8, 25), datetime.date(1028, 8, 24), EraType.GENERAL),
    ("長元", "Chougen", datetime.date(1028, 8, 24), datetime.date(1037, 5, 15), EraType.GENERAL),
    ("長暦", "Chouryaku", datetime.date(1037, 5, 15), datetime.date(1040, 12, 22), EraType.GENERAL),
    ("長久", "Choukyuu", datetime.date(1040, 12, 22), datetime.date(1044, 12, 22), EraType.GENERAL),
    ("寛徳", "Kantoku", datetime.date(1044, 12, 22), datetime.date(1046, 5, 28), EraType.GENERAL),
    ("永承", "Eishou", datetime.date(1046, 5, 28), datetime.date(1053, 2, 8), EraType.GENERAL),
    ("天喜", "Tenki", datetime.date(1053, 2, 8), datetime.date(1058, 9, 25), EraType.GENERAL),
    ("康平", "Kouhei", datetime.date(1058, 9, 25), datetime.date(1065, 9, 10), EraType.GENERAL),
    ("治暦", "Jiryaku", datetime.date(1065, 9, 10), datetime.date(1069, 5, 12), EraType.GENERAL),
    ("延久", "Enkyuu", datetime.date(1069, 5, 12), datetime.date(1074, 9, 22), EraType.GENERAL),
    ("承保", "Jouhou", datetime.date(1074, 9, 22), datetime.date(1077, 12, 11), EraType.GENERAL),
    ("承暦", "Jouryaku", datetime.date(1077, 12, 11), datetime.date(1081, 3, 28), EraType.GENERAL),
    ("永保", "Eihou", datetime.date(1081, 3, 28), datetime.date(1084, 3, 21), EraType.GENERAL),
    ("応徳", "Outoku", datetime.date(1084, 3, 21), datetime.date(1087, 5, 17), EraType.GENERAL),
    ("寛治", "Kanji", datetime.date(1087, 5, 17), datetime.date(1095, 1, 29), EraType.GENERAL),
    ("嘉保", "Kahou", datetime.date(1095, 1, 29), datetime.date(1097, 1, 9), EraType.GENERAL),
    ("永長", "Eichou", datetime.date(1097, 1, 9), datetime.date(1098, 1, 2), EraType.GENERAL),
    ("承徳", "Joutoku", datetime.date(1098, 1, 2), datetime.date(1099, 9, 21), EraType.GENERAL),
    ("康和", "Kouwa", datetime.date(1099, 9, 21), datetime.date(1104, 3, 15), EraType.GENERAL),
    ("長治", "Chouji", datetime.date(1104, 3, 15), datetime.date(1106, 5, 20), EraType.GENERAL),
    ("嘉承", "Kashou", datetime.date(1106, 5, 20), datetime.date(1108, 9, 16), EraType.GENERAL),
    ("天仁", "Tennin", datetime.date(1108, 9, 16), datetime.date(1110, 8, 7), EraType.GENERAL),
    ("天永", "Tennei", datetime.date(1110, 8, 7), datetime.date(1113, 9, 1), EraType.GENERAL),
    ("永久", "Eikyuu", datetime.date(1113, 9, 1), datetime.date(1118, 5, 2), EraType.GENERAL),
    ("元永", "Gennei", datetime.date(1118, 5, 2), datetime.date(1120, 5, 16), EraType.GENERAL),
    ("保安", "Houan", datetime.date(1120, 5, 16), datetime.date(1124, 5, 25), EraType.GENERAL),
    ("天治", "Tenji", datetime.date(1124, 5, 25), datetime.date(1126, 2, 22), EraType.GENERAL),
    ("大治", "Daiji", datetime.date(1126, 2, 22), datetime.date(1131, 3, 7), EraType.GENERAL),
    ("天承", "Tenshou", datetime.date(1131, 3, 7), datetime.date(1132, 9, 28), EraType.GENERAL),
    ("長承", "Choushou", datetime.date(1132, 9, 28), datetime.date(1135, 6, 17), EraType.GENERAL),
    ("保延", "Houen", datetime.date(1135, 6, 17), datetime.date(1141, 8, 20), EraType.GENERAL),
    ("永治", "Eiji", datetime.date(1141, 8, 20), datetime.date(1142, 6, 1), EraType.GENERAL),
    ("康治", "Kouji", datetime.date(1142, 6, 1), datetime.date(1144, 4, 4), EraType.GENERAL),
    ("天養", "Tennyou", datetime.date(1144, 4, 4), datetime.date(1145, 8, 19), EraType.GENERAL),
    ("久安", "Kyuuan", datetime.date(1145, 8, 19), datetime.date(1151, 2, 21), EraType.GENERAL),
    ("仁平", "Ninmpei", datetime.date(1151, 2, 21), datetime.date(1154, 12, 11), EraType.GENERAL),
    ("久寿", "Kyuuju", datetime.date(1154, 12, 11), datetime.date(1156, 5, 25), EraType.GENERAL),
    ("保元", "Hougen", datetime.date(1156, 5, 25), datetime.date(1159, 5, 16), EraType.GENERAL),
    ("平治", "Heiji", datetime.date(1159, 5, 16), datetime.date(1160, 2, 25), EraType.GENERAL),
    ("永暦", "Eiryaku", datetime.date(1160, 2, 25), datetime.date(1161, 10, 1), EraType.GENERAL),
    ("応保", "Ouhou", datetime.date(1161, 10, 1), datetime.date(1163, 5, 11), EraType.GENERAL),
    ("長寛", "Choukan", datetime.date(1163, 5, 11), datetime.date(1165, 7, 21), EraType.GENERAL),
    ("永万", "Eiman", datetime.date(1165, 7, 21), datetime.date(1166, 9, 30), EraType.GENERAL),
    ("仁安", "Ninnan", datetime.date(1166, 9, 30), datetime.date(1169, 5, 13), EraType.GENERAL),
    ("嘉応", "Kaou", datetime.date(1169, 5, 13), datetime.date(1171, 6, 3), EraType.GENERAL),
    ("承安", "Syouan", datetime.date(1171, 6, 3), datetime.date(1175, 8, 23), EraType.GENERAL),
    ("安元", "Angen", datetime.date(1175, 8, 23), datetime.date(1177, 9, 5), EraType.GENERAL),
    ("治承", "Jishou", datetime.date(1177, 9, 5), datetime.date(1181, 9, 1), EraType.GENERAL),
    ("養和", "Youwa", datetime.date(1181, 9, 1), datetime.date(1182, 7, 6), EraType.GENERAL),
    ("寿永", "Juei", datetime.date(1182, 7, 6), datetime.date(1184, 6, 3), EraType.GENERAL),
    ("元暦", "Genryaku", datetime.date(1184, 6, 3), datetime.date(1185, 9, 16), EraType.GENERAL),
    ("文治", "Bunji", datetime.date(1185, 9, 16), datetime.date(1190, 5, 23), EraType.GENERAL),
    ("建久", "Kenkyuu", datetime.date(1190, 5, 23), datetime.date(1199, 5, 30), EraType.GENERAL),
    ("正治", "Syouji", datetime.date(1199, 5, 30), datetime.date(1201, 3, 26), EraType.GENERAL),
    ("建仁", "Kennin", datetime.date(1201, 3, 26), datetime.date(1204, 3, 30), EraType.GENERAL),
    ("元久", "Genkyuu", datetime.date(1204, 3, 30), datetime.date(1206, 6, 12), EraType.GENERAL),
    ("建永", "Kennei", datetime.date(1206, 6, 12), datetime.date(1207, 11, 23), EraType.GENERAL),
    ("承元", "Jougen", datetime.date(1207, 11, 23), datetime.date(1211, 4, 30), EraType.GENERAL),
    ("建暦", "Kenryaku", datetime.date(1211, 4, 30), datetime.date(1214, 1, 25), EraType.GENERAL),
    ("建保", "Kempou", datetime.date(1214, 1, 25), datetime.date(1219, 6, 3), EraType.GENERAL),
    ("承久", "Joukyuu", datetime.date(1219, 6, 3), datetime.date(1222, 6, 1), EraType.GENERAL),
    ("貞応", "Jouou", datetime.date(1222, 6, 1), datetime.date(1225, 1, 7), EraType.GENERAL),
    ("元仁", "Gennin", datetime.date(1225, 1, 7), datetime.date(1225, 6, 4), EraType.GENERAL),
    ("嘉禄", "Karoku", datetime.date(1225, 6, 4), datetime.date(1228, 1, 25), EraType.GENERAL),
    ("安貞", "Antei", datetime.date(1228, 1, 25), datetime.date(1229, 4, 7), EraType.GENERAL),
    ("寛喜", "Kanki", datetime.date(1229, 4, 7), datetime.date(1232, 4, 30), EraType.GENERAL),
    ("貞永", "Jouei", datetime.date(1232, 4, 30), datetime.date(1233, 6, 1), EraType.GENERAL),
    ("天福", "Tempuku", datetime.date(1233, 6, 1), datetime.date(1234, 12, 4), EraType.GENERAL),
    ("文暦", "Bunryaku", datetime.date(1234, 12, 4), datetime.date(1235, 11, 8), EraType.GENERAL),
    ("嘉禎", "Katei", datetime.date(1235, 11, 8), datetime.date(1239, 1, 6), EraType.GENERAL),
    ("暦仁", "Ryakunin", datetime.date(1239, 1, 6), datetime.date(1239, 3, 20), EraType.GENERAL),
    ("延応", "Ennou", datetime.date(1239, 3, 20), datetime.date(1240, 8, 12), EraType.GENERAL),
    ("仁治", "Ninji", datetime.date(1240, 8, 12), datetime.date(1243, 3, 25), EraType.GENERAL),
    ("寛元", "Kangen", datetime.date(1243, 3, 25), datetime.date(1247, 4, 12), EraType.GENERAL),
    ("宝治", "Houji", datetime.date(1247, 4, 12), datetime.date(1249, 5, 9), EraType.GENERAL),
    ("建長", "Kenchou", datetime.date(1249, 5, 9), datetime.date(1256, 10, 31), EraType.GENERAL),
    ("康元", "Kougen", datetime.date(1256, 10, 31), datetime.date(1257, 4, 7), EraType.GENERAL),
    ("正嘉", "Syouka", datetime.date(1257, 4, 7), datetime.date(1259, 4, 27), EraType.GENERAL),
    ("正元", "Syougen", datetime.date(1259, 4, 27), datetime.date(1260, 5, 31), EraType.GENERAL),
    ("文応", "Bunnou", datetime.date(1260, 5, 31), datetime.date(1261, 3, 29), EraType.GENERAL),
    ("弘長", "Kouchou", datetime.date(1261, 3, 29), datetime.date(1264, 4, 3), EraType.GENERAL),
    ("文永", "Bunnei", datetime.date(1264, 4, 3), datetime.date(1275, 5, 29), EraType.GENERAL),
    ("建治", "Kenji", datetime.date(1275, 5, 29), datetime.date(1278, 3, 30), EraType.GENERAL),
    ("弘安", "Kouan", datetime.date(1278, 3, 30), datetime.date(1288, 6, 5), EraType.GENERAL),
    ("正応", "Syouou", datetime.date(1288, 6, 5), datetime.date(1293, 9, 13), EraType.GENERAL),
    ("永仁", "Einin", datetime.date(1293, 9, 13), datetime.date(1299, 6, 1), EraType.GENERAL),
    ("正安", "Syouan", datetime.date(1299, 6, 1), datetime.date(1302, 12, 18), EraType.GENERAL),
    ("乾元", "Kengen", datetime.date(1302, 12, 18), datetime.date(1303, 9, 24), EraType.GENERAL),
    ("嘉元", "Kagen", datetime.date(1303, 9, 24), datetime.date(1307, 1, 26), EraType.GENERAL),
    ("徳治", "Tokuji", datetime.date(1307, 1, 26), datetime.date(1308, 11, 30), EraType.GENERAL),
    ("延慶", "Enkyou", datetime.date(1308, 11, 30), datetime.date(1311, 5, 25), EraType.GENERAL),
    ("応長", "Ouchou", datetime.date(1311, 5, 25), datetime.date(1312, 5, 5), EraType.GENERAL),
    ("正和", "Syouwa", datetime.date(1312, 5, 5), datetime.date(1317, 3, 24), EraType.GENERAL),
    ("文保", "Bumpou", datetime.date(1317, 3, 24), datetime.date(1319, 5, 26), EraType.GENERAL),
    ("元応", "Gennou", datetime.date(1319, 5, 26), datetime.date(1321, 3, 30), EraType.GENERAL),
    ("元亨", "Gennkou", datetime.date(1321, 3, 30), datetime.date(1325, 1, 2), EraType.GENERAL),
    ("正中", "Syouchuu", datetime.date(1325, 1, 2), datetime.date(1326, 6, 5), EraType.GENERAL),
    ("嘉暦", "Karyaku", datetime.date(1326, 6, 5), datetime.date(1329, 9, 30), EraType.GENERAL),
    ("応永", "Ouei", datetime.date(1394, 8, 10), datetime.date(1428, 6, 19), EraType.GENERAL),
    ("正長", "Syouchou", datetime.date(1428, 6, 19), datetime.date(1429, 10, 12), EraType.GENERAL),
    ("永享", "Eikyou", datetime.date(1429, 10, 12), datetime.date(1441, 3, 19), EraType.GENERAL),
    ("嘉吉", "Kakitsu", datetime.date(1441, 3, 19), datetime.date(1444, 3, 3), EraType.GENERAL),
    ("文安", "Bunnann", datetime.date(1444, 3, 3), datetime.date(1449, 8, 25), EraType.GENERAL),
    ("宝徳", "Houtoku", datetime.date(1449, 8, 25), datetime.date(1452, 8, 19), EraType.GENERAL),
    ("享徳", "Kyoutoku", datetime.date(1452, 8, 19), datetime.date(1455, 9, 15), EraType.GENERAL),
    ("康正", "Koushou", datetime.date(1455, 9, 15), datetime.date(1457, 10, 25), EraType.GENERAL),
    ("長禄", "Chouroku", datetime.date(1457, 10, 25), datetime.date(1461, 2, 10), EraType.GENERAL),
    ("寛正", "Kannshou", datetime.date(1461, 2, 10), datetime.date(1466, 3, 23), EraType.GENERAL),
    ("文正", "Bunnshou", datetime.date(1466, 3, 23), datetime.date(1467, 4, 18), EraType.GENERAL),
    ("応仁", "Ouninn", datetime.date(1467, 4, 18), datetime.date(1469, 6, 17), EraType.GENERAL),
    ("文明", "Bunnmei", datetime.date(1469, 6, 17), datetime.date(1487, 8, 18), EraType.GENERAL),
    ("長享", "Choukyou", datetime.date(1487, 8, 18), datetime.date(1489, 9, 25), EraType.GENERAL),
    ("延徳", "Entoku", datetime.date(1489, 9, 25), datetime.date(1492, 8, 21), EraType.GENERAL),
    ("明応", "Meiou", datetime.date(1492, 8, 21), datetime.date(1501, 3, 28), EraType.GENERAL),
    ("文亀", "Bunnki", datetime.date(1501, 3, 28), datetime.date(1504, 3, 26), EraType.GENERAL),
    ("永正", "Eishou", datetime.date(1504, 3, 26), datetime.date(1521, 10, 3), EraType.GENERAL),
    ("大永", "Daiei", datetime.date(1521, 10, 3), datetime.date(1528, 9, 13), EraType.GENERAL),
    ("享禄", "Kyouroku", datetime.date(1528, 9, 13), datetime.date(1532, 9, 8), EraType.GENERAL),
    ("天文", "Tennbunn", datetime.date(1532, 9, 8), datetime.date(1555, 11, 17), EraType.GENERAL),
    ("弘治", "Kouji", datetime.date(1555, 11, 17), datetime.date(1558, 3, 28), EraType.GENERAL),
    ("永禄", "Eiroku", datetime.date(1558, 3, 28), datetime.date(1570, 6, 6), EraType.GENERAL),
    ("元亀", "Gennki", datetime.date(1570, 6, 6), datetime.date(1573, 9, 4), EraType.GENERAL),
    ("天正", "Tennshou", datetime.date(1573, 9, 4), datetime.date(1593, 1, 10), EraType.GENERAL),
    ("文禄", "Bunnroku", datetime.date(1593, 1, 10), datetime.date(1596, 12, 16), EraType.GENERAL),
    ("慶長", "Keichou", datetime.date(1596, 12, 16), datetime.date(1615, 9, 5), EraType.GENERAL),
    ("元和", "Genna", datetime.date(1615, 9, 5), datetime.date(1624, 4, 17), EraType.GENERAL),
    ("寛永", "Kannei", datetime.date(1624, 4, 17), datetime.date(1645, 1, 13), EraType.GENERAL),
    ("正保", "Syouhou", datetime.date(1645, 1, 13), datetime.date(1648, 4, 7), EraType.GENERAL),
    ("慶安", "Keian", datetime.date(1648, 4, 7), datetime.date(1652, 10, 20), EraType.GENERAL),
    ("承応", "Jouou", datetime.date(1652, 10, 20), datetime.date(1655, 5, 18), EraType.GENERAL),
    ("明暦", "Meireki", datetime.date(1655, 5, 18), datetime.date(1658, 8, 21), EraType.GENERAL),
    ("万治", "Manji", datetime.date(1658, 8, 21), datetime.date(1661, 5, 23), EraType.GENERAL),
    ("寛文", "Kannbunn", datetime.date(1661, 5, 23), datetime.date(1673, 10, 30), EraType.GENERAL),
    ("延宝", "Empou", datetime.date(1673, 10, 30), datetime.date(1681, 11, 9), EraType.GENERAL),
    ("天和", "Tenna", datetime.date(1681, 11, 9), datetime.date(1684, 4, 5), EraType.GENERAL),
    ("貞享", "Joukyou", datetime.date(1684, 4, 5), datetime.date(1688, 10, 23), EraType.GENERAL),
    ("元禄", "Genroku", datetime.date(1688, 10, 23), datetime.date(1704, 4, 16), EraType.GENERAL),
    ("宝永", "Houei", datetime.date(1704, 4, 16), datetime.date(1711, 6, 11), EraType.GENERAL),
    ("正徳", "Syoutoku", datetime.date(1711, 6, 11), datetime.date(1716, 8, 9), EraType.GENERAL),
    ("享保", "Kyouhou", datetime.date(1716, 8, 9), datetime.date(1736, 6, 7), EraType.GENERAL),
    ("元文", "Gennbunn", datetime.date(1736, 6, 7), datetime.date(1741, 4, 12), EraType.GENERAL),
    ("寛保", "Kampou", datetime.date(1741, 4, 12), datetime.date(1744, 4, 3), EraType.GENERAL),
    ("延享", "Enkyou", datetime.date(1744, 4, 3), datetime.date(1748, 8, 5), EraType.GENERAL),
    ("寛延", "Kannenn", datetime.date(1748, 8, 5), datetime.date(1751, 12, 14), EraType.GENERAL),
    ("宝暦", "Houreki", datetime.date(1751, 12, 14), datetime.date(1764, 6, 30), EraType.GENERAL),
    ("明和", "Meiwa", datetime.date(1764, 6, 30), datetime.date(1772, 12, 10), EraType.GENERAL),
    ("安永", "Annei", datetime.date(1772, 12, 10), datetime.date(1781, 4, 25), EraType.GENERAL),
    ("天明", "Tennmei", datetime.date(1781, 4, 25), datetime.date(1789, 2, 19), EraType.GENERAL),
    ("寛政", "Kannsei", datetime.date(1789, 2, 19), datetime.date(1801, 3, 19), EraType.GENERAL),
    ("享和", "Kyouwa", datetime.date(1801, 3, 19), datetime.date(1804, 3, 22), EraType.GENERAL),
    ("文化", "Bunnka", datetime.date(1804, 3, 22), datetime.date(1818, 5, 26), EraType.GENERAL),
    ("文政", "Bunnsei", datetime.date(1818, 5, 26), datetime.date(1831, 1, 23), EraType.GENERAL),
    ("天保", "Tenmpou", datetime.date(1831, 1, 23), datetime.date(1845, 1, 9), EraType.GENERAL),
    ("弘化", "Kouka", datetime.date(1845, 1, 9), datetime.date(1848, 4, 1), EraType.GENERAL),
    ("嘉永", "Kaei", datetime.date(1848, 4, 1), datetime.date(1855, 1, 15), EraType.GENERAL),
    ("安政", "Ansei", datetime.date(1855, 1, 15), datetime.date(1860, 4, 8), EraType.GENERAL),
    ("万延", "Mannei", datetime.date(1860, 4, 8), datetime.date(1861, 3, 29), EraType.GENERAL),
    ("文久", "Bunnkyuu", datetime.date(1861, 3, 29), datetime.date(1864, 3, 27), EraType.GENERAL),
    ("元治", "Genji", datetime.date(1864, 3, 27), datetime.date(1865, 5, 1), EraType.GENERAL),
    ("慶応", "Keiou", datetime.date(1865, 5, 1), datetime.date(1868, 10, 23), EraType.GENERAL),
    ("明治", "Meiji", datetime.date(1868, 1, 23), datetime.date(1912, 7, 30), EraType.GENERAL),
    ("大正", "Taishou", datetime.date(1912, 7, 30), datetime.date(1926, 12, 25), EraType.GENERAL),
    ("昭和", "Shouwa", datetime.date(1926, 12, 25), datetime.date(1989, 1, 8), EraType.GENERAL),
    ("平成", "Heisei", datetime.date(1989, 1, 8), datetime.date(2019, 5, 1), EraType.GENERAL),
    ("令和", "Reiwa", datetime.date(2019, 5, 1), None, EraType.GENERAL)
]

_ERA_DATA_DAIKAKUJI = [
    ("元徳", "Gentoku", datetime.date(1329, 9, 30), datetime.date(1331, 9, 19), EraType.DAIKAKUJI),
    ("元弘", "Genkou", datetime.date(1331, 9, 19), datetime.date(1334, 3, 13), EraType.DAIKAKUJI),
    ("建武", "Kenmu", datetime.date(1334, 3, 13), datetime.date(1336, 4, 19), EraType.DAIKAKUJI),
    ("延元", "Engen", datetime.date(1336, 4, 19), datetime.date(1340, 6, 2), EraType.DAIKAKUJI),
    ("興国", "Koukoku", datetime.date(1340, 6, 2), datetime.date(1347, 1, 28), EraType.DAIKAKUJI),
    ("正平", "Syouhei", datetime.date(1347, 1, 28), datetime.date(1370, 8, 24), EraType.DAIKAKUJI),
    ("建徳", "Kentoku", datetime.date(1370, 8, 24), datetime.date(1372, 5, 9), EraType.DAIKAKUJI),
    ("文中", "Bunchuu", datetime.date(1372, 5, 9), datetime.date(1375, 7, 4), EraType.DAIKAKUJI),
    ("天授", "Tenju", datetime.date(1375, 7, 4), datetime.date(1381, 3, 14), EraType.DAIKAKUJI),
    ("弘和", "Kouwa", datetime.date(1381, 3, 14), datetime.date(1384, 5, 26), EraType.DAIKAKUJI),
    ("元中", "Genchuu", datetime.date(1384, 5, 26), datetime.date(1392, 11, 27), EraType.DAIKAKUJI)
]

_ERA_DATA_JIMYOUIN = [
    ("元徳", "Gentoku", datetime.date(1329, 9, 30), datetime.date(1332, 5, 31), EraType.JIMYOUIN),
    ("正慶", "Shoukyou", datetime.date(1332, 5, 31), datetime.date(1333, 7, 15), EraType.JIMYOUIN),
    ("建武", "Kenmu", datetime.date(1334, 3, 13), datetime.date(1338, 10, 19), EraType.JIMYOUIN),
    ("暦応", "Ryakuou", datetime.date(1338, 10, 19), datetime.date(1342, 6, 9), EraType.JIMYOUIN),
    ("康永", "Kouei", datetime.date(1342, 6, 9), datetime.date(1345, 11, 23), EraType.JIMYOUIN),
    ("貞和", "Jouwa", datetime.date(1345, 11, 23), datetime.date(1350, 4, 12), EraType.JIMYOUIN),
    ("観応", "Kannou", datetime.date(1350, 4, 12), datetime.date(1352, 11, 12), EraType.JIMYOUIN),
    ("文和", "Bunna", datetime.date(1352, 11, 12), datetime.date(1356, 5, 7), EraType.JIMYOUIN),
    ("延文", "Enbun", datetime.date(1356, 5, 7), datetime.date(1361, 5, 12), EraType.JIMYOUIN),
    ("康安", "Kouan", datetime.date(1361, 5, 12), datetime.date(1362, 10, 19), EraType.JIMYOUIN),
    ("貞治", "Jouji", datetime.date(1362, 10, 19), datetime.date(1368, 3, 15), EraType.JIMYOUIN),
    ("応安", "Ouan", datetime.date(1368, 3, 15), datetime.date(1375, 4, 6), EraType.JIMYOUIN),
    ("永和", "Eiwa", datetime.date(1375, 4, 6), datetime.date(1379, 4, 17), EraType.JIMYOUIN),
    ("康暦", "Kouryaku", datetime.date(1379, 4, 17), datetime.date(1381, 3, 28), EraType.JIMYOUIN),
    ("永徳", "Eitoku", datetime.date(1381, 3, 28), datetime.date(1384, 3, 27), EraType.JIMYOUIN),
    ("至徳", "Sitoku", datetime.date(1384, 3, 27), datetime.date(1387, 10, 13), EraType.JIMYOUIN),
    ("嘉慶", "Kakyou", datetime.date(1387, 10, 13), datetime.date(1389, 3, 15), EraType.JIMYOUIN),
    ("康応", "Kouou", datetime.date(1389, 3, 15), datetime.date(1390, 4, 20), EraType.JIMYOUIN),
    ("明徳", "Meitoku", datetime.date(1390, 4, 20), datetime.date(1394, 8, 10), EraType.JIMYOUIN)
]



"""
    Japanera
    -----------

    Easy japanese era tool
    All Information's source is [Wikipedia Page](https://ja.wikipedia.org/wiki/%E5%85%83%E5%8F%B7%E4%B8%80%E8%A6%A7_(%E6%97%A5%E6%9C%AC))
    Powered by [Yamato Nagata](https://twitter.com/514YJ)

    [GitHub](https://github.com/nagataaaas/Japanera)
    [ReadTheDocs](https://japanera.readthedocs.io/en/latest/)

    :copyright: (c) 2019-2024 by Yamato Nagata.
    :license: MIT.
"""

from .__about__ import __version__
from .japanera import (Era, EraDate, EraDateTime, ERA_DATA_COMMON, ERA_DATA_DAIKAKUJI, ERA_DATA_JIMYOUIN,
                       ERA_DATA_GENERAL)
from .era_data import (EraType)

__all__ = [
    __version__,
    "Era",
    "EraDate",
    "EraDateTime",
    "EraType",
    "ERA_DATA_COMMON",
    "ERA_DATA_DAIKAKUJI",
    "ERA_DATA_JIMYOUIN",
    "ERA_DATA_GENERAL",
]



import datetime
import unittest
from datetime import date

from japanera import EraDate, Era, EraType, EraDateTime, ERA_DATA_GENERAL, ERA_DATA_COMMON


class TestEraDate(unittest.TestCase):
    def test_new(self):
        self.assertWarns(RuntimeWarning, EraDate, 2019, 1, 1,
                         Era("Test", "test", date(2000, 1, 1), date(2000, 1, 2), EraType.GENERAL))

    def test_strptime_normal(self):
        result = EraDate.strptime("2020-01-01", "%Y-%m-%d")
        self.assertEqual(len(result), 2)
        self.assertSetEqual(set(result),
                            {EraDate(2020, 1, 1, ERA_DATA_GENERAL[-1]), EraDate(2020, 1, 1, ERA_DATA_COMMON[0])})

        result = EraDate.strptime("300-01-01", "%-Y-%m-%d")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDate(300, 1, 1, ERA_DATA_COMMON[0]))

    def test_strptime_era_name(self):
        result = EraDate.strptime("令和-05-01", "%-K-%m-%d")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDate(2019, 5, 1, ERA_DATA_GENERAL[-1]))

        result = EraDate.strptime("令和-04-01", "%-K-%m-%d")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDate(2020, 4, 1, ERA_DATA_GENERAL[-1]))

    def test_strptime_kanji_date(self):
        result = EraDate.strptime("令和 五月三十一日", "%-K %-m月%-d日")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDate(2019, 5, 31, ERA_DATA_GENERAL[-1]))

        result = EraDate.strptime("令和三年 4月1日", "%-K%-n年 %-m月%-d日")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDate(2021, 4, 1, ERA_DATA_GENERAL[-1]))
        self.assertRaises(ValueError, EraDate.strptime, "令和元年 四月三十一日",
                          "%-K%-n年 %-m月%-d日")  # invalid date and year combination

        result = EraDate.strptime("令和三年 十二月二十四日", "%-K%-n年 %-m月%-d日")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDate(2021, 12, 24, ERA_DATA_GENERAL[-1]))

        result = EraDate.strptime("西暦2023年 1月1日", "%-K%-Y年 %-m月%-d日")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDate(2023, 1, 1, ERA_DATA_COMMON[0]))

    def test_strftime(self):
        era_date = EraDate(1950, 12, 24, ERA_DATA_GENERAL[-3])  # 昭和
        self.assertEqual(era_date.strftime("%-K(%-E, %-e, %-h)%-n年 %-m月%-d日 %H:%M:%S"),
                         "昭和(Shouwa, Showa, S)二十五年 十二月二十四日 00:00:00")

        era_date = EraDate(1926, 12, 24, ERA_DATA_GENERAL[-3])
        self.assertEqual(era_date.strftime("%-K(%-E, %-e, %-h)%-n年 %-m月%-d日(%-a) %H:%M:%S"),
                         "昭和(Shouwa, Showa, S)元年 十二月二十四日(金) 00:00:00")

    def test_from_date(self):
        self.assertEqual(EraDate.from_date(date(300, 1, 1)).era, ERA_DATA_COMMON[0])
        self.assertEqual(EraDate.from_date(date(2300, 1, 1)).era, ERA_DATA_GENERAL[-1])
        self.assertEqual(EraDate.from_date(date(2300, 1, 1), ERA_DATA_COMMON[0]).era, ERA_DATA_COMMON[0])

    def test_to_date(self):
        self.assertEqual(EraDate.from_date(date(300, 1, 1)).to_date(), date(300, 1, 1))
        self.assertEqual(EraDate.from_date(date(2300, 1, 1)).to_date(), date(2300, 1, 1))
        self.assertEqual(EraDate.from_date(date(2300, 1, 1), ERA_DATA_COMMON[0]).to_date(), date(2300, 1, 1))


class TestEraDateTime(unittest.TestCase):
    def test_new(self):
        self.assertWarns(RuntimeWarning, EraDateTime, 2019, 1, 1,
                         era=Era("Test", "test", date(2000, 1, 1), date(2000, 1, 2), EraType.GENERAL))
        self.assertEqual(EraDateTime(2019, 1, 1, era=ERA_DATA_COMMON[0]).era, ERA_DATA_COMMON[0])
        self.assertEqual(EraDateTime(2019, 1, 1).era, ERA_DATA_GENERAL[-2])

    def test_strptime(self):
        result = EraDateTime.strptime("令和-05-01 00:00:00", "%-K-%m-%d %H:%M:%S")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDateTime(2019, 5, 1, 0, 0, 0, era=ERA_DATA_GENERAL[-1]))

        result = EraDateTime.strptime("令和-04-01 00:00:00", "%-K-%m-%d %H:%M:%S")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDateTime(2020, 4, 1, 0, 0, 0, era=ERA_DATA_GENERAL[-1]))

        result = EraDateTime.strptime("令和-04-01 12:34:56.12345", "%-K-%m-%d %H:%M:%S.%f")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDateTime(2020, 4, 1, 12, 34, 56, 123450, era=ERA_DATA_GENERAL[-1]))

        result = EraDateTime.strptime("昭和45年 1/1", "%-K%-y年 %m/%d")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDateTime(1970, 1, 1, era=ERA_DATA_GENERAL[-3]))
        result = EraDateTime.strptime("昭和45年 1/2 UTC+0930", "%-K%-y年 %m/%d %Z%z")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0],
                         EraDateTime(1970, 1, 2,
                                     tzinfo=datetime.timezone(
                                         datetime.timedelta(seconds=9 * 3600 + 30 * 60)),
                                     era=ERA_DATA_GENERAL[-3]))
        self.assertEqual(result[0].timestamp(),
                         datetime.datetime.strptime("1970 1/2 UTC+0930", "%Y %m/%d %Z%z").timestamp())
        result = EraDateTime.strptime("昭和45年 1/1 GMT", "%-K%-y年 %m/%d %Z")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], EraDateTime(1970, 1, 1, era=ERA_DATA_GENERAL[-3]))

    def test_strftime(self):
        era_date = EraDateTime(1950, 12, 24, 12, 34, 56, era=ERA_DATA_GENERAL[-3])
        self.assertEqual(era_date.strftime("%-K(%-E, %-e, %-h)%-n年 %-m月%-d日 %H:%M:%S"),
                         "昭和(Shouwa, Showa, S)二十五年 十二月二十四日 12:34:56")

        era_date = EraDateTime(1926, 12, 24, 12, 34, 56, era=ERA_DATA_GENERAL[-3])
        self.assertEqual(era_date.strftime("%-K(%-E, %-e, %-h)%-n年 %-m月%-d日(%-a) %H:%M:%S"),
                         "昭和(Shouwa, Showa, S)元年 十二月二十四日(金) 12:34:56")


if __name__ == '__main__':
    unittest.main()



__title__ = "Japanera"
__description__ = "Easy japanese era tool"
__url__ = "https://github.com/nagataaaas/Japanera"
__version_info__ = ("2", "1", "1")
__version__ = ".".join(__version_info__)
__author__ = "Yamato Nagata"
__author_email__ = "nagata@nagata.pro"
__maintainer__ = "Yamato Nagata"
__license__ = "MIT"
__copyright__ = "(c) 2019-2024 by Yamato Nagata"



import unittest
from datetime import date, timedelta

import kanjize

from japanera import parser, era_data, Era, ERA_DATA_COMMON, ERA_DATA_JIMYOUIN, ERA_DATA_DAIKAKUJI, ERA_DATA_GENERAL


class TestStrPTime(unittest.TestCase):
    def test_vanilla_use_timezone(self):
        (era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year), \
        (year, month, day, hour, minute, second, weekday, julian, tz, tzname, gmtoff), \
        fraction, gmtoff_fraction = parser._strptime("2020-10-23(Fri) 11:30:42.345678:UTC",
                                                     "%Y-%m-%d(%a) %H:%M:%S.%f:%Z")
        self.assertEqual([(era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year),
                          (year, month, day, hour, minute, second, weekday, julian, tz, tzname,
                           gmtoff), fraction, gmtoff_fraction],
                         [(None, None, None, None, None),
                          (2020, 10, 23, 11, 30, 42, 4, None, 0, 'UTC', None),
                          345678, 0])

    def test_vanilla_use_offset(self):
        (era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year), \
        (year, month, day, hour, minute, second, weekday, julian, tz, tzname, gmtoff), \
        fraction, gmtoff_fraction = parser._strptime("2020-10-23(Sun) 11:30:42.345678:+0900",
                                                     "%Y-%m-%d(%a) %H:%M:%S.%f:%z")
        self.assertEqual([(era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year),
                          (year, month, day, hour, minute, second, weekday, julian, tz, tzname,
                           gmtoff), fraction, gmtoff_fraction],
                         [(None, None, None, None, None),
                          (2020, 10, 23, 11, 30, 42, 6, None, -1, None, 32400),
                          345678, 0])

    def test_era_name_parse(self):
        (era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year), \
        (year, month, day, hour, minute, second, weekday, julian, tz, tzname, gmtoff), \
        fraction, gmtoff_fraction = parser._strptime("宝暦(Houreki, Horeki, H)", "%-K(%-E, %-e, %-h)")
        self.assertEqual([(era_kanji, era_english, era_english_vowel_shortened, era_head, relative_year),
                          (year, month, day, hour, minute, second, weekday, julian, tz, tzname,
                           gmtoff), fraction, gmtoff_fraction],
                         [("宝暦", "Houreki", "Horeki", "H", None),
                          (None, None, None, 0, 0, 0, None, None, -1, None, None),
                          0, 0])

    def test_era_name_parse_error(self):
        self.assertRaises(ValueError, parser._strptime, "玲和", "%-K")
        self.assertRaises(ValueError, parser._strptime, "Rewa", "%-E")
        self.assertRaises(ValueError, parser._strptime, "Rewa", "%-e")
        self.assertRaises(ValueError, parser._strptime, "Q", "%-h")

    def test_parse_kanji_date(self):
        self.assertEqual(parser._strptime("2000", "%-Y")[1][0], 2000)
        self.assertEqual(parser._strptime("二千", "%-N")[1][0], 2000)
        self.assertEqual(parser._strptime("二千百四十六", "%-N")[1][0], 2146)
        self.assertEqual(parser._strptime("千十", "%-N")[1][0], 1010)
        self.assertEqual(parser._strptime("四百九十八", "%-N")[1][0], 498)
        self.assertEqual(parser._strptime("元", "%-N")[1][0], 1)
        self.assertEqual(parser._strptime("元", "%-Y")[1][0], 1)
        self.assertRaises(ValueError, parser._strptime, "四万", "%-N")  # out of range
        self.assertRaises(ValueError, parser._strptime, "4千", "%-N")  # invalid format
        self.assertRaises(ValueError, parser._strptime, "4000", "%-N")  # invalid format
        self.assertRaises(ValueError, parser._strptime, "四千", "%-Y")  # invalid format

        self.assertEqual(parser._strptime("20", "%-y")[0][4], 20)
        self.assertEqual(parser._strptime("三十八", "%-n")[0][4], 38)
        self.assertEqual(parser._strptime("十一", "%-n")[0][4], 11)
        self.assertEqual(parser._strptime("十", "%-n")[0][4], 10)
        self.assertEqual(parser._strptime("元", "%-y")[0][4], 1)
        self.assertEqual(parser._strptime("元", "%-n")[0][4], 1)
        self.assertRaises(ValueError, parser._strptime, "二千", "%-n")  # out of range
        self.assertRaises(ValueError, parser._strptime, "五百二十一", "%-n")  # out of range
        self.assertRaises(ValueError, parser._strptime, "五百二十一", "%-y")  # out of range
        self.assertRaises(ValueError, parser._strptime, "50", "%-n")  # invalid format
        self.assertRaises(ValueError, parser._strptime, "二十", "%-y")  # invalid format

        for month in range(1, 13):
            self.assertEqual(parser._strptime(kanjize.number2kanji(month), "%-m")[1][1], month)
            self.assertEqual(parser._strptime(str(month), "%-m")[1][1], month)
        self.assertRaises(ValueError, parser._strptime, "十三", "%-m")  # out of range
        self.assertRaises(ValueError, parser._strptime, "元", "%-m")  # invalid character

        for day in range(1, 32):
            self.assertEqual(parser._strptime(kanjize.number2kanji(day), "%-d")[1][2], day)
            self.assertEqual(parser._strptime(str(day), "%-d")[1][2], day)
        self.assertRaises(ValueError, parser._strptime, "三十二", "%-d")  # out of range
        self.assertRaises(ValueError, parser._strptime, "元", "%-d")  # invalid character

        today = date.today()
        for offset in range(7):
            target = today + timedelta(days=offset)
            self.assertEqual(parser._strptime('月火水木金土日'[target.weekday()], "%-a")[1][6],
                             parser._strptime(target.strftime("%a"), "%a")[1][6])
            self.assertEqual('月火水木金土日'[target.weekday()],
                             {0: '月', 1: '火', 2: '水', 3: '木', 4: '金', 5: '土',
                              6: '日'}[parser._strptime(target.strftime("%a"), "%a")[1][6]])  # check weekday assignment
        self.assertRaises(ValueError, parser._strptime, "元", "%-a")  # invalid character
        self.assertRaises(ValueError, parser._strptime, "天", "%-a")  # invalid character


class TestFindClosestLeapYear(unittest.TestCase):
    def test(self):
        self.assertEqual(parser.find_closest_leap_year(2000), 2000)
        self.assertEqual(parser.find_closest_leap_year(1999), 2000)
        self.assertEqual(parser.find_closest_leap_year(2100), 2104)
        self.assertEqual(parser.find_closest_leap_year(2099), 2104)
        self.assertEqual(parser.find_closest_leap_year(2039), 2040)
        self.assertEqual(parser.find_closest_leap_year(2040), 2040)
        self.assertEqual(parser.find_closest_leap_year(2041), 2044)


class TestFindErasWithYear(unittest.TestCase):
    def test_only_general(self):
        self.assertSetEqual(parser.find_eras_with_year(1000),
                            {
                                Era("長保", "Chouhou", date(999, 2, 6),
                                           date(1004, 8, 14), era_data.EraType.GENERAL),
                                Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                            })

    def test_both_nambokucho(self):
        self.assertSetEqual(parser.find_eras_with_year(1331),
                            {
                                Era("元徳", "Gentoku", date(1329, 9, 30),
                                           date(1331, 9, 19), era_data.EraType.DAIKAKUJI),
                                Era("元弘", "Genkou", date(1331, 9, 19),
                                           date(1334, 3, 13), era_data.EraType.DAIKAKUJI),
                                Era("元徳", "Gentoku", date(1329, 9, 30),
                                           date(1332, 5, 31), era_data.EraType.JIMYOUIN),
                                Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                            })

    def test_nambokucho_only_jimyouin(self):
        self.assertSetEqual(parser.find_eras_with_year(1393),
                            {
                                Era("明徳", "Meitoku", date(1390, 4, 20),
                                           date(1394, 8, 10), era_data.EraType.JIMYOUIN),
                                Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                            })

    def test_before_era(self):
        self.assertSetEqual(parser.find_eras_with_year(1),
                            {
                                Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON)
                            })

    def test_latest(self):
        self.assertSetEqual(parser.find_eras_with_year(3000),
                            {
                                ERA_DATA_GENERAL[-1],
                                Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                            })


class TestFindEraAndDate(unittest.TestCase):
    def test_only_kanji(self):
        self.assertListEqual(parser.find_era_and_date(era_kanji="令和"),
                             [(Era("令和", "Reiwa", date(2019, 5, 1), None, era_data.EraType.GENERAL),
                               date(2019, 5, 1))])
        self.assertListEqual(parser.find_era_and_date(era_kanji="元徳"),
                             [
                                 (Era("元徳", "Gentoku", date(1329, 9, 30),
                                             date(1331, 9, 19), era_data.EraType.DAIKAKUJI),
                                  date(1329, 9, 30)),
                                 (Era("元徳", "Gentoku", date(1329, 9, 30),
                                             date(1332, 5, 31), era_data.EraType.JIMYOUIN),
                                  date(1329, 9, 30)),
                             ])
        self.assertRaises(ValueError, parser.find_era_and_date, era_kanji="無効")

    def test_only_english(self):
        self.assertListEqual(parser.find_era_and_date(era_english="Reiwa"),
                             [(Era("令和", "Reiwa", date(2019, 5, 1), None, era_data.EraType.GENERAL),
                               date(2019, 5, 1))])
        self.assertListEqual(parser.find_era_and_date(era_english="Kouwa"),
                             [
                                 (Era("康和", "Kouwa", date(1099, 9, 21),
                                             date(1104, 3, 15), era_data.EraType.GENERAL),
                                  date(1099, 9, 21)),
                                 (Era("弘和", "Kouwa", date(1381, 3, 14),
                                             date(1384, 5, 26), era_data.EraType.DAIKAKUJI),
                                  date(1381, 3, 14)),
                             ])
        self.assertRaises(ValueError, parser.find_era_and_date, era_english="Invalid")

    def test_only_english_vowel_shortened(self):
        self.assertListEqual(parser.find_era_and_date(era_english_vowel_shortened="Tengyo"),
                             [(Era("天慶", "Tengyou", date(938, 6, 27), date(947, 5, 20),
                                          era_data.EraType.GENERAL),
                               date(938, 6, 27))])
        self.assertRaises(ValueError, parser.find_era_and_date, era_english_vowel_shortened="Tengyou")  # not shortened
        self.assertRaises(ValueError, parser.find_era_and_date, era_english_vowel_shortened="Invalid")

    def test_only_english_head(self):
        self.assertListEqual(parser.find_era_and_date(era_head_english="W"),
                             [(Era("和銅", "Wadou", date(708, 2, 11), date(715, 10, 7),
                                          era_data.EraType.GENERAL),
                               date(708, 2, 11))])
        self.assertRaises(ValueError, parser.find_era_and_date, era_english_vowel_shortened="Wadou")  # not shortened
        self.assertRaises(ValueError, parser.find_era_and_date, era_english_vowel_shortened="Q")  # no era starts with Q

    def test_only_absolute_year(self):
        self.assertListEqual(parser.find_era_and_date(absolute_year=710),
                             [(Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                               date(710, 1, 1)),
                              (Era("和銅", "Wadou", date(708, 2, 11), date(715, 10, 7),
                                          era_data.EraType.GENERAL),
                               date(710, 1, 1))])
        self.assertListEqual(parser.find_era_and_date(absolute_year=715),  # 2 eras
                             [(Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                               date(715, 1, 1)),
                              (Era("和銅", "Wadou", date(708, 2, 11), date(715, 10, 7),
                                          era_data.EraType.GENERAL),
                               date(715, 1, 1)),
                              (Era("霊亀", "Reiki", date(715, 10, 7), date(717, 12, 28),
                                          era_data.EraType.GENERAL),
                               date(715, 10, 7)),
                              ])
        self.assertListEqual(parser.find_era_and_date(absolute_year=1225),  # 3 eras
                             [(Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                               date(1225, 1, 1)),
                              (Era("貞応", "Jouou", date(1222, 6, 1), date(1225, 1, 7),
                                          era_data.EraType.GENERAL),
                               date(1225, 1, 1)),
                              (Era("元仁", "Gennin", date(1225, 1, 7), date(1225, 6, 4),
                                          era_data.EraType.GENERAL),
                               date(1225, 1, 7)),
                              (Era("嘉禄", "Karoku", date(1225, 6, 4), date(1228, 1, 25),
                                          era_data.EraType.GENERAL),
                               date(1225, 6, 4)),
                              ])
        self.assertListEqual(parser.find_era_and_date(absolute_year=2500),
                             [(Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                               date(2500, 1, 1)),
                              (ERA_DATA_GENERAL[-1],
                               date(2500, 1, 1)),
                              ])
        self.assertListEqual(parser.find_era_and_date(absolute_year=1),
                             [(Era("西暦", "Seireki", date(1, 1, 1), None, era_data.EraType.COMMON),
                               date(1, 1, 1)),
                              ])

    def test_only_relative_year(self):
        self.assertEqual(len(parser.find_era_and_date(relative_year=1)),
                         len(ERA_DATA_GENERAL + ERA_DATA_JIMYOUIN +
                             ERA_DATA_DAIKAKUJI + ERA_DATA_COMMON))
        self.assertListEqual([e.kanji for e, _ in parser.find_era_and_date(relative_year=50)],
                             ["西暦", "昭和", "令和"])
        self.assertListEqual([e.kanji for e, _ in parser.find_era_and_date(relative_year=100)],
                             ["西暦", "令和"])

    def test_combination_era_name(self):
        self.assertListEqual(parser.find_era_and_date(era_kanji="康和", era_english="Kouwa"),
                             [
                                 (Era("康和", "Kouwa", date(1099, 9, 21),
                                             date(1104, 3, 15), era_data.EraType.GENERAL),
                                  date(1099, 9, 21)),
                             ])

        self.assertListEqual(
            parser.find_era_and_date(era_english="Kouwa", era_english_vowel_shortened="Kowa", era_head_english="K"),
            [
                (Era("康和", "Kouwa", date(1099, 9, 21),
                            date(1104, 3, 15), era_data.EraType.GENERAL),
                 date(1099, 9, 21)),
                (Era("弘和", "Kouwa", date(1381, 3, 14),
                            date(1384, 5, 26), era_data.EraType.DAIKAKUJI),
                 date(1381, 3, 14)),
            ])

    def test_combination_era_name_year_and_date(self):
        self.assertListEqual(parser.find_era_and_date(era_english="Kouwa", relative_year=4),
                             [
                                 (Era("康和", "Kouwa", date(1099, 9, 21),
                                             date(1104, 3, 15), era_data.EraType.GENERAL),
                                  date(1102, 1, 1)),
                                 (Era("弘和", "Kouwa", date(1381, 3, 14),
                                             date(1384, 5, 26), era_data.EraType.DAIKAKUJI),
                                  date(1384, 1, 1)),
                             ])
        self.assertListEqual(parser.find_era_and_date(era_english="Kouwa", relative_year=5),
                             [
                                 (Era("康和", "Kouwa", date(1099, 9, 21),
                                             date(1104, 3, 15), era_data.EraType.GENERAL),
                                  date(1103, 1, 1)),
                             ])

        self.assertListEqual(parser.find_era_and_date(era_english="Kouwa", relative_year=4, month=5, day=25),
                             [
                                 (Era("康和", "Kouwa", date(1099, 9, 21),
                                             date(1104, 3, 15), era_data.EraType.GENERAL),
                                  date(1102, 5, 25)),
                                 (Era("弘和", "Kouwa", date(1381, 3, 14),
                                             date(1384, 5, 26), era_data.EraType.DAIKAKUJI),
                                  date(1384, 5, 25)),
                             ])
        self.assertListEqual(parser.find_era_and_date(era_english="Kouwa", relative_year=4, month=5, day=26),
                             [
                                 (Era("康和", "Kouwa", date(1099, 9, 21),
                                             date(1104, 3, 15), era_data.EraType.GENERAL),
                                  date(1102, 5, 26)),
                             ])
        self.assertRaises(ValueError, parser.find_era_and_date, era_english="Kouwa", absolute_year=1500)

    def test_date(self):
        self.assertEqual(parser.find_era_and_date(era_kanji="大正")[0][1], date(1912, 7, 30))
        self.assertEqual(parser.find_era_and_date(era_kanji="大正", relative_year=1)[0][1], date(1912, 7, 30))
        self.assertEqual(parser.find_era_and_date(era_kanji="大正", relative_year=2)[0][1], date(1913, 1, 1))
        self.assertEqual(parser.find_era_and_date(era_kanji="大正", month=8)[0][1], date(1912, 8, 1))
        self.assertEqual(parser.find_era_and_date(era_kanji="大正", month=1)[0][1], date(1913, 1, 1))
        self.assertEqual(parser.find_era_and_date(era_kanji="大正", month=2, day=29)[0][1], date(1916, 2, 29))
        self.assertEqual(len(parser.find_era_and_date(absolute_year=2019, month=4, day=30)), 2)

    def test_allow_date_after_end_of_era(self):
        self.assertListEqual(parser.find_era_and_date(era_kanji="平成", relative_year=31, month=4, day=30),
                             [
                                 (Era("平成", "Heisei", date(1989, 1, 8),
                                             date(2019, 5, 1), era_data.EraType.GENERAL),
                                  date(2019, 4, 30)),
                             ])
        self.assertListEqual(parser.find_era_and_date(era_kanji="平成", relative_year=31, month=5, day=1), [])
        self.assertListEqual(parser.find_era_and_date(era_kanji="平成", relative_year=31, month=5, day=1,
                                                      allow_date_after_end_of_era=True),
                             [
                                 (Era("平成", "Heisei", date(1989, 1, 8),
                                             date(2019, 5, 1), era_data.EraType.GENERAL),
                                  date(2019, 5, 1)),
                             ])


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import date

from japanera import Era, EraType


class TestEra(unittest.TestCase):
    def test_compare_era(self):
        before_era = Era("前", "Mae", date(2000, 1, 1), date(2001, 1, 1), EraType.GENERAL)
        between_era = Era("間", "Aida", date(2001, 5, 1), date(2001, 6, 1), EraType.GENERAL)
        after_era = Era("後", "Ato", date(2001, 1, 1), date(2002, 1, 1), EraType.GENERAL)

        before_before = date(1999, 12, 31)
        start_before = date(2000, 1, 1)
        in_before = date(2000, 6, 1)
        last_before = date(2001, 1, 1)

        self.assertTrue(before_before < before_era)
        self.assertTrue(before_era > before_before)
        self.assertFalse(start_before < before_era)
        self.assertTrue(start_before <= before_era)
        self.assertTrue(in_before in before_era)
        self.assertFalse(last_before in before_era)
        self.assertTrue(last_before not in before_era)
        self.assertFalse(last_before < before_era)
        self.assertFalse(last_before <= before_era)
        self.assertTrue(last_before > before_era)
        self.assertTrue(last_before >= before_era)

        self.assertTrue(in_before in before_era)

        self.assertTrue(before_era < after_era)
        self.assertTrue(before_era <= after_era)
        self.assertTrue(before_era <= between_era)
        self.assertTrue(between_era >= after_era)


if __name__ == '__main__':
    unittest.main()



"""
Easy japanese era tool
----------------------
Powered by [Yamato Nagata](https://twitter.com/514YJ)

[GitHub](https://github.com/delta114514/Japanera)
[ReadTheDocs](https://japanera.readthedocs.io/en/latest/)


```python
>>> from datetime import date
>>> from japanera import EraDate

>>> today = EraDate.from_date(date.today())
>>> date(2020, 4, 16) in today.era
True

>>> "Current Japanese Era is <{}>: <{}>".format(today.era.kanji, today.era.english)
Current Japanese Era is <令和>: <Reiwa>

>>> "Current Date is <{}>".format(today.strftime("%-K%-y年%m月%d日"))
Current Date is <令和05年03月07日>
```
"""

from os import path

from setuptools import setup

about = {}
with open("japanera/__about__.py") as f:
    exec(f.read(), about)

here = path.abspath(path.dirname(__file__))

setup(name=about["__title__"],
      version=about["__version__"],
      url=about["__url__"],
      license=about["__license__"],
      author=about["__author__"],
      author_email=about["__author_email__"],
      description=about["__description__"],
      long_description=__doc__,
      long_description_content_type="text/markdown",
      install_requires=["kanjize==1.4.0"],
      packages=["japanera"],
      zip_safe=False,
      setup_requires=['wheel'],
      platforms="any",
      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: Other Environment",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python",
          "Topic :: Software Development :: Libraries :: Python Modules"
      ])
